"""
Created 2021

@author: wf
"""

import csv
import io
from typing import Any, Dict, List

from lodstorage.lod import LOD


class CSV:
    """
    helper for converting data in csv format to list of dicts (LoD) and vice versa
    """

    _instance = None

    def __new__(cls, dialect: str = "excel", quoting: int = csv.QUOTE_NONNUMERIC):
        """
        constructor to set dialect and quoting defaults
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.dialect = dialect
            cls._instance.quoting = quoting
        return cls._instance

    @classmethod
    def get_instance(cls, dialect: str = "excel", quoting: int = csv.QUOTE_NONNUMERIC):
        """Get singleton instance of CSV class"""
        return cls(dialect, quoting)

    def restoreFromCSVFile(
        self, filePath: str, headerNames: List[str] = None, withPostfix: bool = False
    ) -> List[Dict[str, Any]]:
        """
        restore LOD from given csv file

        Args:
            filePath(str): file name
            headerNames(List[str]): Names of the headers that should be used. If None it is assumed that the header is given.
            withPostfix(bool): If False the file type is appended to given filePath. Otherwise file type MUST be given with filePath.

        Returns:
            List[Dict[str, Any]]: list of dicts (LoD) containing the content of the given csv file
        """
        if not withPostfix:
            filePath += ".csv"
        csvStr = self.readFile(filePath)
        lod = self.fromCSV(csvStr, headerNames)
        return lod

    def fromCSV(
        self,
        csvString: str,
        fields: List[str] = None,
        dialect: str = None,
        quoting: int = None,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        convert given csv string to list of dicts (LOD)

        Args:
            csvString(str): csv string that should be converted to LOD
            fields(List[str]): Names of the headers that should be used. If None it is assumed that the header is given.
            dialect(str): CSV dialect to use
            quoting(int): CSV quoting behavior

        Returns:
            List[Dict[str, Any]]: list of dicts (LoD) containing the content of the given csv string
        """
        if dialect is None:
            dialect = self.dialect
        if quoting is None:
            quoting = self.quoting
        csvStream = io.StringIO(csvString)
        reader = csv.DictReader(
            csvStream, fieldnames=fields, dialect=dialect, quoting=quoting, **kwargs
        )
        lod = list(reader)
        self.fixTypes(lod)
        return lod

    def storeToCSVFile(
        self, lod: List[Dict[str, Any]], filePath: str, withPostfix: bool = False
    ):
        """
        converts the given lod to CSV file.

        Args:
            lod(List[Dict[str, Any]]): lod that should be converted to csv file
            filePath(str): file name the csv should be stored to
            withPostfix(bool): If False the file type is appended to given filePath. Otherwise file type MUST be given with filePath.
        """
        if not withPostfix:
            filePath += ".csv"
        csvStr = self.toCSV(lod)
        self.writeFile(csvStr, filePath)

    def toCSV(
        self,
        lod: List[Dict[str, Any]],
        includeFields: List[str] = None,
        excludeFields: List[str] = None,
        dialect: str = None,
        quoting: int = None,
        **kwargs,
    ) -> str:
        """
        converts the given lod to CSV string.

        Args:
            lod(List[Dict[str, Any]]): lod that should be converted to csv string
            includeFields(List[str]): list of fields that should be included in the csv
            excludeFields(List[str]): list of fields that should be excluded from the csv
            dialect(str): CSV dialect to use
            quoting(int): CSV quoting behavior

        Returns:
            str: csv string of the given lod
        """
        if dialect is None:
            dialect = self.dialect
        if quoting is None:
            quoting = self.quoting
        if lod is None:
            return ""
        if excludeFields is not None:
            lod = LOD.filterFields(lod, excludeFields)
        if includeFields is None:
            fields = LOD.getFields(lod)
        else:
            fields = includeFields
            lod = LOD.filterFields(lod, includeFields, reverse=True)
        csvStream = io.StringIO()
        dict_writer = csv.DictWriter(
            csvStream, fieldnames=fields, dialect=dialect, quoting=quoting, **kwargs
        )
        dict_writer.writeheader()
        dict_writer.writerows(lod)
        csvString = csvStream.getvalue()
        return csvString

    def readFile(self, filename: str) -> str:
        """
        Reads the given filename and returns it as string

        Args:
            filename(str): Name of the file that should be returned as string

        Returns:
            str: Content of the file as string
        """
        with open(filename, "r") as file:
            content = file.read()
        return content

    def writeFile(self, content: str, filename: str):
        """
        Write the given str to the given filename

        Args:
            content(str): string that should be written into the file
            filename(str): Name of the file the given str should be written to
        """
        with open(filename, "w") as file:
            file.write(content)

    def fixTypes(self, lod: List[Dict[str, Any]]) -> None:
        """
        fixes the types of the given LoD.

        Args:
            lod(List[Dict[str, Any]]): List of dictionaries to fix types for
        """
        for record in lod:
            for key, value in record.items():
                if value == "":
                    record[key] = None



"""
Created on 2025-06-01

@author: wf
"""

from enum import Enum

from SPARQLWrapper import JSON, N3, RDF, TURTLE


class RdfFormat(Enum):
    """
    RDF serialization formats with MIME types, file extensions, and SPARQLWrapper constants.
    """

    TURTLE = ("turtle", "text/turtle", ".ttl", TURTLE)
    RDF_XML = ("rdf-xml", "application/rdf+xml", ".rdf", RDF)
    N3 = ("n3", "text/n3", ".n3", N3)
    JSON_LD = ("json-ld", "application/ld+json", ".jsonld", JSON)

    def __init__(self, label: str, mime_type: str, extension: str, sparql_format):
        self.label = label
        self.mime_type = mime_type
        self.extension = extension
        self.sparql_format = sparql_format

    @classmethod
    def by_label(cls, label: str):
        """Get format by label"""
        for rdf_format in cls:
            if rdf_format.label == label:
                return rdf_format
        raise ValueError(f"Unknown format: {label}")



"""
Created on 2021-06-13

@author: wf
"""

from collections import Counter

from tabulate import tabulate

from lodstorage.sample2 import Sample
from lodstorage.tabulateCounter import TabulateCounter
from tests.basetest import Basetest


class TestTabulate(Basetest):
    """
    test tabulate support/compatibility
    """

    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def testIssue24_IntegrateTabulate(self):
        """
        https://github.com/WolfgangFahl/pyLoDStorage/issues/24

        test https://pypi.org/project/tabulate/ support
        """
        show = self.debug
        # show = True
        royals_lod = Sample.getRoyals()
        for fmt in ["latex", "grid", "mediawiki", "github"]:
            table = tabulate(royals_lod, headers="keys", tablefmt=fmt)
            if show:
                print(table)
            self.assertTrue("Charles III" in table)

        cities = Sample.getCities()
        counter = Counter()
        for city in cities:
            counter[city["country"]] += 1
        tabulateCounter = TabulateCounter(counter)
        for fmt in ["latex", "grid", "mediawiki", "github"]:
            table = tabulateCounter.mostCommonTable(tablefmt=fmt, limit=7)
            if show:
                print(table)
        pass



"""
Created on 2024-03-02

@author: wf
"""

import json

from lodstorage.prefixes import Prefixes
from tests.basetest import Basetest


class TestPrefixes(Basetest):
    """
    test SPARQL prefixes
    """

    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.sparql_query = """PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX wd: <http://www.wikidata.org/entity/>
PREFIX wdt: <http://www.wikidata.org/prop/direct/>

SELECT ?item ?label WHERE {
  ?item wdt:P31 wd:Q5 .
  ?item rdfs:label ?label .
  FILTER(LANG(?label) = "en")
}
"""

    def test_prefixes(self):
        """
        test the SPARQL prefix providing
        """
        sparql_code = Prefixes.getPrefixes(prefixes=["rdf", "xsd"])
        expected = """PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
"""
        self.assertEqual(expected, sparql_code)

    def test_extract_prefixes(self):
        """
        Test the SPARQL prefix extraction functionality.
        """
        prefixes = Prefixes.extract_prefixes(self.sparql_query)
        if self.debug:
            print(json.dumps(prefixes, indent=2))
        expected = ["rdf", "rdfs", "wd", "wdt"]
        for prefix in expected:
            self.assertTrue(prefix in prefixes)
            self.assertEqual(prefixes[prefix], Prefixes.prefixMap[prefix])

    def test_merge_prefixes(self):
        """
        Test merging prefixes into SPARQL query.
        """
        prefixes = Prefixes.getPrefixes(prefixes=["rdf", "xsd"])
        merged_query = Prefixes.merge_prefixes(self.sparql_query, prefixes)
        if self.debug:
            print(merged_query)
        expected = Prefixes.getPrefixes(["xsd"])
        self.assertTrue(merged_query.startswith(expected))



"""
Created on 2025-06-04

@author: wf
"""

import json
from pathlib import Path
from typing import Dict

from lodstorage.prefix_config import PrefixConfigs
from lodstorage.prefixes import Prefixes
from lodstorage.query import EndpointManager
from tests.basetest import Basetest


class TestPrefixConfig(Basetest):
    """
    Test Prefix Configuration
    """

    def setUp(self, debug=False, profile=True):
        """
        setUp the test environment
        """
        Basetest.setUp(self, debug=debug, profile=profile)
        self.prefixes_yaml_path = (
            Path(__file__).parent / ".." / "sampledata" / "prefixes.yaml"
        ).resolve()
        self.pfix_configs = PrefixConfigs.ofYaml(self.prefixes_yaml_path)

    def get_all_prefixes(self) -> Dict[str, str]:
        """
        Get all prefixes from all prefix sets combined

        Returns:
            Dict[str, str]: Combined dictionary of all prefix mappings
        """
        all_prefixes = {}
        for prefix_config in self.pfix_configs.prefix_sets.values():
            all_prefixes.update(prefix_config.prefixes)
        return all_prefixes

    def test_load_prefixes(self):
        """
        test loading prefix configurations
        """
        pfix_configs = self.pfix_configs
        if self.debug:
            for set_name, pfix_config in pfix_configs.prefix_sets.items():
                print(f"{set_name}:{pfix_config.description}")
                print(json.dumps(pfix_config.prefixes, indent=2, default=str))
        self.assertTrue("wikidata" in pfix_configs.prefix_sets)
        self.assertTrue("wdt" in pfix_configs.prefix_sets["wikidata"].prefixes)

    def test_complete(self):
        """
        test we have references all PREFIX definitions of all default endpoints
        """
        endpoints = EndpointManager.getEndpoints(lang="sparql")
        all_config_prefixes = self.get_all_prefixes()
        missing_prefixes = set()

        for name, endpoint in endpoints.items():

            if hasattr(endpoint, "prefixes") and endpoint.prefixes:
                if self.debug:
                    pf_list = endpoint.prefixes.split("\n")
                    print(f"checking {len(pf_list)} prefixes of endpoint {name}")
                endpoint_prefixes = Prefixes.extract_prefixes(endpoint.prefixes)

                for prefix_name, prefix_uri in endpoint_prefixes.items():
                    if prefix_name not in all_config_prefixes:
                        missing_prefixes.add(f"{prefix_name}: {prefix_uri}")
                        if self.debug:
                            print(f"  Missing prefix: {prefix_name} -> {prefix_uri}")
        self.assertEqual(
            len(missing_prefixes),
            0,
            f"Missing {len(missing_prefixes)} prefixes: {missing_prefixes}",
        )

    def test_get_declarations(self):
        """
        test getting declarations works
        """
        prefix_sets = ["rdf", "blazegraph", "gov"]
        declarations = self.pfix_configs.get_selected_declarations(prefix_sets)

        if self.debug:
            print(f"Selected declarations for {prefix_sets}:")
            print(declarations)
        expected_prefixes = ["rdf", "bd", "gov"]
        for prefix in expected_prefixes:
            self.assertIn(prefix, declarations)



"""
Created on 2024-03-02

@author: wf
"""

import re


class Prefixes:
    """
    Handles the generation of standard SPARQL prefix declarations for queries.
    This utility class simplifies the inclusion of common prefixes used in SPARQL
    queries by providing a method to generate the necessary PREFIX lines based on
    a list of prefix keys.

    The class supports a wide range of prefixes relevant to Wikidata and general RDF/SPARQL
    usage, including RDF, RDFS, Wikibase, Schema.org, and more. It aims to reduce redundancy
    and improve clarity in SPARQL query construction by centralizing prefix management.

    Attributes:
        None

    Methods:
        getPrefixes(prefixes): Generates SPARQL PREFIX lines for a given list of prefix keys.
    """

    prefixMap = {
        "bd": "<http://www.bigdata.com/rdf#>",
        "cc": "<http://creativecommons.org/ns#>",
        "dct": "<http://purl.org/dc/terms/>",
        "geo": "<http://www.opengis.net/ont/geosparql#>",
        "ontolex": "<http://www.w3.org/ns/lemon/ontolex#>",
        "owl": "<http://www.w3.org/2002/07/owl#>",
        "p": "<http://www.wikidata.org/prop/>",
        "pq": "<http://www.wikidata.org/prop/qualifier/>",
        "pqn": "<http://www.wikidata.org/prop/qualifier/value-normalized/>",
        "pqv": "<http://www.wikidata.org/prop/qualifier/value/>",
        "pr": "<http://www.wikidata.org/prop/reference/>",
        "prn": "<http://www.wikidata.org/prop/reference/value-normalized/>",
        "prov": "<http://www.w3.org/ns/prov#>",
        "prv": "<http://www.wikidata.org/prop/reference/value/>",
        "ps": "<http://www.wikidata.org/prop/statement/>",
        "psn": "<http://www.wikidata.org/prop/statement/value-normalized/>",
        "psv": "<http://www.wikidata.org/prop/statement/value/>",
        "rdf": "<http://www.w3.org/1999/02/22-rdf-syntax-ns#>",
        "rdfs": "<http://www.w3.org/2000/01/rdf-schema#>",
        "schema": "<http://schema.org/>",
        "skos": "<http://www.w3.org/2004/02/skos/core#>",
        "wd": "<http://www.wikidata.org/entity/>",
        "wdata": "<http://www.wikidata.org/wiki/Special:EntityData/>",
        "wdno": "<http://www.wikidata.org/prop/novalue/>",
        "wdref": "<http://www.wikidata.org/reference/>",
        "wds": "<http://www.wikidata.org/entity/statement/>",
        "wdt": "<http://www.wikidata.org/prop/direct/>",
        "wdtn": "<http://www.wikidata.org/prop/direct-normalized/>",
        "wdv": "<http://www.wikidata.org/value/>",
        "wikibase": "<http://wikiba.se/ontology#>",
        "xsd": "<http://www.w3.org/2001/XMLSchema#>",
    }

    @classmethod
    def getPrefixes(
        cls, prefixes=["rdf", "rdfs", "schema", "wd", "wdt", "wikibase", "xsd"]
    ) -> str:
        """Generates SPARQL PREFIX lines for a given list of prefix keys.

        This method looks up URIs for the specified prefixes from a predefined map and constructs
        PREFIX lines suitable for inclusion at the beginning of a SPARQL query. It allows for easy
        and flexible specification of the prefixes needed for a particular query.

        Args:
            prefixes (list of str): A list of prefix keys for which PREFIX lines should be generated.
                Defaults to a common set of prefixes used in Wikidata queries.

        Returns:
            str: A string containing the SPARQL PREFIX lines for the specified prefixes, each ending
                with a newline character. If a prefix key is not recognized, it is ignored.

        Example:
            >>> Prefixes.getPrefixes(["wd", "wdt"])
            'PREFIX wd: <http://www.wikidata.org/entity/>\nPREFIX wdt: <http://www.wikidata.org/prop/direct/>\n'
        """

        # see also https://www.wikidata.org/wiki/EntitySchema:E49
        prefixes = cls.prefix_string(cls.prefixMap, prefixes)
        return prefixes

    @classmethod
    def prefix_string(cls, prefix_dict: dict, prefix_keys: list[str]):
        prefixes = ""
        for prefix in prefix_keys:
            if prefix in prefix_dict:
                prefixes += cls.prefix_line(prefix_dict, prefix)
        return prefixes

    @classmethod
    def prefix_line(cls, prefix_dict: dict, prefix: str) -> str:
        line = f"PREFIX {prefix}: {prefix_dict[prefix]}\n"
        return line

    @classmethod
    def extract_prefixes(cls, sparql_query: str) -> dict:
        """
        Extract only the explicitly declared prefixes from a SPARQL query string.
        Simple regex-based extraction that finds PREFIX declarations in the query text.

        Args:
            sparql_query (str): The SPARQL query containing PREFIX declarations

        Returns:
            dict: Dictionary mapping prefix names to their URI strings
        """
        declared_prefixes = {}

        # Simple pattern to match PREFIX declarations: PREFIX name: <uri>
        prefix_pattern = r"PREFIX\s+(\w+):\s*<([^>]+)>"

        # Find all PREFIX declarations (case insensitive)
        matches = re.findall(prefix_pattern, sparql_query, re.IGNORECASE)

        # Convert matches to dictionary
        for prefix_name, uri in matches:
            declared_prefixes[prefix_name] = f"<{uri}>"

        return declared_prefixes

    @classmethod
    def merge_prefix_dict(cls, query: str, prefix_dict: dict) -> str:
        """
        Merge prefixes from dict into SPARQL query by prepending missing prefix declarations.

        Args:
            query (str): The SPARQL query
            prefix_dict (dict): Dictionary of prefixes to merge

        Returns:
            str: SPARQL query with missing prefixes prepended
        """
        existing_prefixes = cls.extract_prefixes(query)
        missing = set(prefix_dict.keys()) - set(existing_prefixes.keys())
        prepend = cls.prefix_string(prefix_dict, list(missing))
        query = prepend + query

        return query

    @classmethod
    def merge_prefixes(cls, query: str, prefixes: str) -> str:
        """
        Merge prefixes from string into SPARQL query by prepending missing prefix declarations.

        Args:
            query (str): The SPARQL query
            prefixes (str): String containing PREFIX declarations

        Returns:
            str: SPARQL query with missing prefixes prepended
        """
        prefix_dict = cls.extract_prefixes(prefixes)
        merged_query = cls.merge_prefix_dict(query, prefix_dict)
        return merged_query



"""
Created on 2025-06-12

@author: wf
"""

import os

from lodstorage.params import Param
from lodstorage.query import Query, QueryManager
from tests.basetest import Basetest


class TestQueryYaml(Basetest):
    """
    Test query yaml serialization handling
    """

    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.wikidata_queries_path = (
            f"{os.path.dirname(__file__)}/../sampledata/wikidata.yaml"
        )

    def testRead(self):
        """
        test reading a yaml file
        """
        qm = QueryManager(
            queriesPath=self.wikidata_queries_path,
            with_default=False,
            lang="sparql",
            debug=False,
        )
        self.assertTrue(len(qm.queriesByName) > 8)
        pass

    def testQuerySerialization(self):
        """
        Test YAML serialization of Query objects with parameters
        """
        debug = self.debug
        debug = True

        # Create a Query with parameters manually
        input_param = Param(name="qids", type="str", description="Wikidata QIDs")
        output_param = Param(name="label", type="str", description="Entity label")

        query = Query(
            name="TestQuery",
            query="SELECT ?entity ?label WHERE { VALUES ?entity { {{qids}} } ?entity rdfs:label ?label }",
            param_list=[input_param],
            output=[output_param],
        )

        # Test serialization
        yaml_str = query.to_yaml()

        if debug:
            print("Query YAML:")
            print(yaml_str)

        # Should not contain Python object references
        self.assertNotIn("!!python/object:", yaml_str)

        # Test deserialization
        restored_query = Query.from_yaml(yaml_str)  # @UndefinedVariable
        self.assertEqual(restored_query.name, "TestQuery")
        self.assertEqual(len(restored_query.param_list), 1)
        self.assertEqual(len(restored_query.output), 1)



"""
Created on 2025-06-04

@author: wf
"""

from dataclasses import dataclass, field
from typing import Dict, List, Optional

from basemkit.yamlable import lod_storable


@dataclass
class PrefixConfig:
    """
    Configuration for SPARQL prefixes
    """

    name: str
    description: Optional[str] = None
    prefixes: Dict[str, str] = field(default_factory=dict)

    def as_text(self) -> str:
        """
        Get prefixes as formatted text block.

        Returns:
            str: Newline-separated prefix declarations
        """
        text = "\n".join(self.prefixes.values())
        return text


@lod_storable
class PrefixConfigs:
    """Collection of prefix configurations loaded from YAML."""

    prefix_sets: Dict[str, PrefixConfig] = field(default_factory=dict)

    @classmethod
    def ofYaml(cls, yaml_path: str) -> "PrefixConfigs":
        """Load prefix configurations from YAML file."""
        prefix_configs = cls.load_from_yaml_file(yaml_path)
        return prefix_configs

    def __post_init__(self):
        """ """
        self.all_prefixes = {}
        for prefix_config in self.prefix_sets.values():
            self.all_prefixes.update(prefix_config.prefixes)
        pass

    def get_selected_declarations(self, prefix_set: List[str]) -> str:
        """
        Get PREFIX declarations for selected prefix sets.

        Args:
            prefix_set: List of prefix set names to include

        Returns:
            str: Combined PREFIX declarations
        """
        selected_prefixes = {}
        for prefix_set_name in prefix_set:
            if prefix_set_name in self.prefix_sets:
                prefix_config = self.prefix_sets[prefix_set_name]
                selected_prefixes.update(prefix_config.prefixes)
        declarations = self.get_prefix_declarations(selected_prefixes)
        return declarations

    def get_prefix_declarations(self, prefixes: Dict[str, str] = None) -> str:
        """
        Convert prefixes to PREFIX declarations.

        Args:
            prefixes: Dictionary of prefix mappings, defaults to all_prefixes

        Returns:
            str: Newline-separated PREFIX declarations
        """
        if prefixes is None:
            prefixes = self.all_prefixes
        prefix_lines = []
        for prefix_name, prefix_uri in prefixes.items():
            prefix_lines.append(f"PREFIX {prefix_name}: <{prefix_uri}>")
            declarations = "\n".join(prefix_lines)
        return declarations



"""
Created 2021

@author: wf
"""

import csv
import tempfile

from lodstorage.lod import LOD
from lodstorage.lod_csv import CSV
from lodstorage.sample2 import Sample
from tests.basetest import Basetest


class TestCSV(Basetest):
    """
    Tests functionalities for the conversion between csv and list od dicts (LoD)
    """

    def setUp(self):
        super().setUp(debug=False)
        self.csv_converter = CSV.get_instance()
        self.temp_dir = tempfile.TemporaryDirectory(prefix="test_pyLoDStorage_")
        self.testFolder = self.temp_dir.name
        self.csvStr = '"pageTitle","name","label"\r\n"page_1","Test Page 1","1"\r\n"page_2","Test Page 2","2"\r\n'
        self.csvStrMinimal = (
            "pageTitle,name,label\r\npage_1,Test Page 1,1\r\npage_2,Test Page 2,2\r\n"
        )
        self.csvLOD = [
            {"pageTitle": "page_1", "name": "Test Page 1", "label": "1"},
            {"pageTitle": "page_2", "name": "Test Page 2", "label": "2"},
        ]
        self.quoting_configs = [(csv.QUOTE_MINIMAL, self.csvStrMinimal)]

    def tearDown(self) -> None:
        super().tearDown()
        self.temp_dir.cleanup()

    def testRoyals(self):
        """test conversion of royals with different quoting modes"""
        inlod = Sample.getRoyals()
        for quoting, _ in self.quoting_configs:
            with self.subTest(quoting=quoting):
                csv_output = self.csv_converter.toCSV(inlod, quoting=quoting)
                if self.debug:
                    print(f"Quoting {quoting}: {csv_output}")
                outlod = self.csv_converter.fromCSV(csv_output, quoting=quoting)
                if self.debug:
                    print(f"Parsed: {outlod}")

    def test_to_csv(self):
        """tests if LoD is correctly converted to csv str with various quoting"""
        for quoting, expected in self.quoting_configs:
            with self.subTest(quoting=quoting):
                actual = self.csv_converter.toCSV(self.csvLOD, quoting=quoting)
                self.assertEqual(expected, actual)

    def test_to_csv_incomplete_dicts(self):
        """tests if LoD is correctly converted to csv even if some dicts are incomplete"""
        csvLOD = [
            {"pageTitle": "page_1", "label": "1"},
            {"pageTitle": "page_2", "name": "Test Page 2", "label": "2"},
        ]
        expected_cases = [
            (
                csv.QUOTE_MINIMAL,
                "pageTitle,label,name\r\npage_1,1,\r\npage_2,2,Test Page 2\r\n",
            )
        ]
        for quoting, expected in expected_cases:
            with self.subTest(quoting=quoting):
                actual = self.csv_converter.toCSV(csvLOD, quoting=quoting)
                self.assertEqual(expected, actual)

    def test_store_to_csvfile(self):
        """tests if LoD is correctly stored as csv file"""
        for quoting, expected_content in self.quoting_configs:
            with self.subTest(quoting=quoting):
                fileName = f"{self.testFolder}/test_store_{quoting}.csv"
                expected = expected_content.replace("\r", "")
                csvStr = self.csv_converter.toCSV(self.csvLOD, quoting=quoting)
                self.csv_converter.writeFile(csvStr, fileName)
                actual = self.csv_converter.readFile(fileName)
                self.assertEqual(expected, actual)

    def test_from_csv(self):
        """tests if csv is correctly parsed to LoD"""
        for quoting, csv_content in self.quoting_configs:
            with self.subTest(quoting=quoting):
                lod = self.csv_converter.fromCSV(csv_content, quoting=quoting)
                self.assertEqual(len(lod), 2)
                for key, value in self.csvLOD[0].items():
                    self.assertIn(key, lod[0])
                    self.assertEqual(value, lod[0][key])

    def test_restore_from_csvfile(self):
        """tests if lod is correctly restored from csv file"""
        for quoting, csv_content in self.quoting_configs:
            with self.subTest(quoting=quoting):
                fileName = f"{self.testFolder}/test_restore_{quoting}.csv"
                self.csv_converter.writeFile(csv_content, fileName)
                lod = self.csv_converter.fromCSV(
                    self.csv_converter.readFile(fileName), quoting=quoting
                )
                self.assertEqual(self.csvLOD, lod)

    def test_round_trip(self):
        """tests csv round trip: dict -> csv -> dict"""
        csvLOD = [
            {"pageTitle": "page_1", "name": "Test Page 1", "label": "1"},
            {"name": "Test Page 2", "label": "2"},
            {"pageTitle": "page_3", "label": "3"},
            {"pageTitle": "page_4", "name": "Test Page 4"},
        ]
        for quoting, _ in self.quoting_configs:
            with self.subTest(quoting=quoting):
                fileName = f"{self.testFolder}/round_trip_{quoting}.csv"
                csvStr = self.csv_converter.toCSV(csvLOD, quoting=quoting)
                self.csv_converter.writeFile(csvStr, fileName)
                actualLOD = self.csv_converter.fromCSV(
                    self.csv_converter.readFile(fileName), quoting=quoting
                )
                expectedLOD = csvLOD.copy()
                fields = LOD.getFields(expectedLOD)
                LOD.setNone4List(expectedLOD, fields)
                self.assertEqual(expectedLOD, actualLOD)

    def test_from_csv_without_header(self):
        """tests if csv string without embedded headers is parsed correctly"""
        test_cases = [
            (csv.QUOTE_MINIMAL, "page_1,Test Page 1,1\r\npage_2,Test Page 2,2\r\n")
        ]
        headerNames = ["pageTitle", "name", "label"]
        for quoting, csv_content in test_cases:
            with self.subTest(quoting=quoting):
                actualLOD = self.csv_converter.fromCSV(
                    csv_content, headerNames, quoting=quoting
                )
                self.assertEqual(self.csvLOD, actualLOD)

    def test_to_csv_delimiter_in_value(self):
        """tests if delimiter in dict value will not result in incorrect values"""
        csvLOD = [
            {
                "pageTitle": "page_1",
                "name": "Test Page 1, delimiter in value",
                "label": "1,000",
            }
        ]
        for quoting, _ in self.quoting_configs:
            with self.subTest(quoting=quoting):
                actualCSVStr = self.csv_converter.toCSV(csvLOD, quoting=quoting)
                actualLOD = self.csv_converter.fromCSV(actualCSVStr, quoting=quoting)
                self.assertEqual(csvLOD, actualLOD)
